import ast
import json
import tempfile
from pathlib import Path
from textwrap import dedent
from typing import Any, List
from unittest.mock import Mock, patch

import keycloak.exceptions
import pytest
import requests.exceptions
import yaml
from typer.testing import CliRunner

from _nebari.cli import create_cli
from _nebari.keycloak import get_expected_roles

TEST_KEYCLOAK_USERS = [
    {"id": "1", "username": "test-dev", "groups": ["analyst", "developer"]},
    {"id": "2", "username": "test-admin", "groups": ["admin"]},
    {"id": "3", "username": "test-nogroup", "groups": []},
]

TEST_KEYCLOAK_GROUPS = [
    {"name": "admin", "path": "/admin"},
    {"name": "analyst", "path": "/analyst"},
    {"name": "developer", "path": "/developer"},
    {"name": "superadmin", "path": "/superadmin"},
    {"name": "users", "path": "/users"},
]

TEST_KEYCLOAK_GROUP_CLIENT_ROLES = {
    "/admin": {
        "realm-management": ["query-users", "query-groups", "manage-users"],
        "jupyterhub": [
            "jupyterhub_admin",
            "dask_gateway_admin",
            "allow-group-directory-creation-role",
        ],
        "grafana": ["grafana_admin"],
        "argo-server-sso": ["argo-admin"],
        "conda_store": ["conda_store_admin"],
    },
    "/analyst": {
        "jupyterhub": [
            "jupyterhub_developer",
            "allow-group-directory-creation-role",
            "allow-read-access-to-services-role",
        ],
        "grafana": ["grafana_viewer"],
        "argo-server-sso": ["argo-viewer"],
        "conda_store": ["conda_store_developer"],
    },
    "/developer": {
        "jupyterhub": [
            "jupyterhub_developer",
            "allow-group-directory-creation-role",
            "dask_gateway_developer",
        ],
        "grafana": ["grafana_developer"],
        "argo-server-sso": ["argo-developer"],
        "conda_store": ["conda_store_developer"],
    },
    "/superadmin": {
        "realm-management": ["realm-admin"],
        "conda_store": ["conda_store_superadmin"],
    },
    "/users": {},
}


TEST_DOMAIN = "nebari.example.com"
MOCK_KEYCLOAK_ENV = {
    "KEYCLOAK_SERVER_URL": f"https://{TEST_DOMAIN}/auth/",
    "KEYCLOAK_ADMIN_USERNAME": "root",
    "KEYCLOAK_ADMIN_PASSWORD": "super-secret-123!",
}

TEST_ACCESS_TOKEN = "abc123"

runner = CliRunner()


def parse_table(table_text, headers: list):
    """
    A simple parser that extracts the rows from your table into a list of tuples/dicts.
    Assumes that each row is in the format and order of the headers list.
    """
    rows = []
    # Split by lines and filter out lines that don't have '│' columns
    for line in table_text.splitlines():
        if line.strip().startswith("│"):
            # Example split by '│', ignoring first and last segments
            columns = [col.strip() for col in line.split("│")[1:-1]]
            if len(columns) == len(headers):
                rows.append(dict(zip(headers, columns)))
    return rows


def parse_yaml(stdout_str: str, headers: list):
    """
    A simple parser that extracts the rows from your YAML into a list of dicts.
    Assumes your YAML is a list of objects, and each object might contain fields
    corresponding to the given headers list.
    """
    # Use yaml.safe_load to parse the YAML string into Python data
    data = yaml.safe_load(stdout_str)

    # If the YAML is empty or not a list, ensure we handle gracefully
    if not data or not isinstance(data, list):
        return []

    parsed_rows = []

    # Loop over each object in the YAML list
    for item in data:
        # Build a new dict containing only the desired headers in the specified order
        row = {}
        for h in headers:
            row[h] = item.get(h, None)  # or some default if the key is missing
        parsed_rows.append(row)

    return parsed_rows


def test_parse_table():
    table_str = dedent(
        """                      Keycloak Users (Count: 3)
        ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ Username     ┃ Email                    ┃ Groups                   ┃
        ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ test-dev     │ test-dev@example.com     │ ['analyst', 'developer'] │
        │ test-admin   │ test-admin@example.com   │ ['admin']                │
        │ test-nogroup │ test-nogroup@example.com │ []                       │
        └──────────────┴──────────────────────────┴──────────────────────────┘
        """
    )
    expected = [
        {
            "username": "test-dev",
            "email": "test-dev@example.com",
            "groups": "['analyst', 'developer']",
        },
        {
            "username": "test-admin",
            "email": "test-admin@example.com",
            "groups": "['admin']",
        },
        {
            "username": "test-nogroup",
            "email": "test-nogroup@example.com",
            "groups": "[]",
        },
    ]
    parsed_result = parse_table(table_str, ["username", "email", "groups"])
    print(parsed_result)
    assert 3 == len(parsed_result)
    for i, row in enumerate(parsed_result):
        assert expected[i] == row


def test_parse_yaml():
    yaml_str = dedent(
        """
        - name: test
          roles:
            - allow-test-to-pass
            - other-role
        """
    )
    expected = [
        {
            "name": "test",
            "roles": ["allow-test-to-pass", "other-role"],
        },
    ]

    parsed_result = parse_yaml(yaml_str, ["name", "roles"])
    print(parsed_result)
    assert 1 == len(parsed_result)
    assert expected[0] == parsed_result[0]


@pytest.mark.parametrize(
    "args, exit_code, content",
    [
        # --help
        ([], 0, ["Usage:"]),
        (["--help"], 0, ["Usage:"]),
        (["-h"], 0, ["Usage:"]),
        (["add-user", "--help"], 0, ["Usage:"]),
        (["add-user", "-h"], 0, ["Usage:"]),
        (["export-users", "--help"], 0, ["Usage:"]),
        (["export-users", "-h"], 0, ["Usage:"]),
        (["list-users", "--help"], 0, ["Usage:"]),
        (["list-users", "-h"], 0, ["Usage:"]),
        (["list-groups", "--help"], 0, ["Usage:"]),
        (["list-groups", "-h"], 0, ["Usage:"]),
        # error, missing args
        (["add-user"], 2, ["Missing option"]),
        (["add-user", "--config"], 2, ["requires an argument"]),
        (["add-user", "-c"], 2, ["requires an argument"]),
        (["add-user", "--user"], 2, ["requires an argument"]),
        (["export-users"], 2, ["Missing option"]),
        (["export-users", "--config"], 2, ["requires an argument"]),
        (["export-users", "-c"], 2, ["requires an argument"]),
        (["export-users", "--realm"], 2, ["requires an argument"]),
        (["list-users"], 2, ["Missing option"]),
        (["list-users", "--config"], 2, ["requires an argument"]),
        (["list-users", "-c"], 2, ["requires an argument"]),
        (["list-groups"], 2, ["Missing option"]),
        (["list-groups", "--config"], 2, ["requires an argument"]),
        (["list-groups", "-c"], 2, ["requires an argument"]),
    ],
)
def test_cli_keycloak_stdout(args: List[str], exit_code: int, content: List[str]):
    app = create_cli()
    result = runner.invoke(app, ["keycloak"] + args)
    assert result.exit_code == exit_code
    for c in content:
        assert c in result.stdout


@patch("keycloak.KeycloakAdmin")
def test_cli_keycloak_adduser_happy_path_from_env(_mock_keycloak_admin):
    result = run_cli_keycloak_adduser(use_env=True)

    assert 0 == result.exit_code
    assert not result.exception
    assert f"Created user={TEST_KEYCLOAK_USERS[0]['username']}" in result.stdout


@patch("keycloak.KeycloakAdmin")
def test_cli_keycloak_adduser_happy_path_from_config(_mock_keycloak_admin):
    result = run_cli_keycloak_adduser(use_env=False)

    assert 0 == result.exit_code
    assert not result.exception
    assert f"Created user={TEST_KEYCLOAK_USERS[0]['username']}" in result.stdout


@patch(
    "keycloak.KeycloakAdmin.__init__",
    side_effect=keycloak.exceptions.KeycloakConnectionError(
        error_message="connection test"
    ),
)
def test_cli_keycloak_adduser_keycloak_connection_exception(_mock_keycloak_admin):
    result = run_cli_keycloak_adduser()

    assert 1 == result.exit_code
    assert result.exception
    assert "Failed to connect to Keycloak server: connection test" in str(
        result.exception
    )


@patch(
    "keycloak.KeycloakAdmin.__init__",
    side_effect=keycloak.exceptions.KeycloakAuthenticationError(
        error_message="auth test"
    ),
)
def test_cli_keycloak_adduser_keycloak_auth_exception(_mock_keycloak_admin):
    result = run_cli_keycloak_adduser()

    assert 1 == result.exit_code
    assert result.exception
    assert "Failed to connect to Keycloak server: auth test" in str(result.exception)


@patch(
    "keycloak.KeycloakAdmin",
    return_value=Mock(
        create_user=Mock(
            side_effect=keycloak.exceptions.KeycloakConnectionError(
                error_message="unhandled"
            )
        ),
    ),
)
def test_cli_keycloak_adduser_keycloak_unhandled_error(_mock_keycloak_admin):
    result = run_cli_keycloak_adduser()

    assert 1 == result.exit_code
    assert result.exception
    assert "unhandled" == str(result.exception)


@patch(
    "keycloak.KeycloakAdmin",
    return_value=Mock(
        users_count=Mock(side_effect=lambda: len(TEST_KEYCLOAK_USERS)),
        get_users=Mock(
            side_effect=lambda: [
                {
                    "id": u["id"],
                    "username": u["username"],
                    "email": f"{u['username']}@example.com",
                }
                for u in TEST_KEYCLOAK_USERS
            ]
        ),
        get_user_groups=Mock(
            side_effect=lambda user_id: [
                {"name": g}
                for u in TEST_KEYCLOAK_USERS
                if u["id"] == user_id
                for g in u["groups"]
            ]
        ),
    ),
)
def test_cli_keycloak_listusers_happy_path_from_env(_mock_keycloak_admin):
    result = run_cli_keycloak_listusers(use_env=True)

    assert 0 == result.exit_code
    assert not result.exception

    parsed_result = parse_table(result.stdout, ["username", "email", "groups"])

    # output should start with the number of users found then
    # display a table with their info
    assert len(TEST_KEYCLOAK_USERS) == len(parsed_result)

    for i, u in enumerate(TEST_KEYCLOAK_USERS):
        assert u["username"] == parsed_result[i]["username"]
        assert f"{u['username']}@example.com" == parsed_result[i]["email"]
        assert u["groups"] == ast.literal_eval(parsed_result[i]["groups"])


@patch(
    "keycloak.KeycloakAdmin",
    return_value=Mock(
        users_count=Mock(side_effect=lambda: len(TEST_KEYCLOAK_USERS)),
        get_users=Mock(
            side_effect=lambda: [
                {
                    "id": u["id"],
                    "username": u["username"],
                    "email": f"{u['username']}@example.com",
                }
                for u in TEST_KEYCLOAK_USERS
            ]
        ),
        get_user_groups=Mock(
            side_effect=lambda user_id: [
                {"name": g}
                for u in TEST_KEYCLOAK_USERS
                if u["id"] == user_id
                for g in u["groups"]
            ]
        ),
    ),
)
def test_cli_keycloak_listusers_happy_path_from_config(_mock_keycloak_admin):
    result = run_cli_keycloak_listusers(use_env=False)

    assert 0 == result.exit_code
    assert not result.exception

    parsed_result = parse_table(result.stdout, ["username", "email", "groups"])

    # output should start with the number of users found then
    # display a table with their info
    assert len(TEST_KEYCLOAK_USERS) == len(parsed_result)

    for i, u in enumerate(TEST_KEYCLOAK_USERS):
        assert u["username"] == parsed_result[i]["username"]
        assert f"{u['username']}@example.com" == parsed_result[i]["email"]
        assert u["groups"] == ast.literal_eval(parsed_result[i]["groups"])


@patch(
    "keycloak.KeycloakAdmin",
    return_value=Mock(
        get_groups=Mock(
            side_effect=lambda: TEST_KEYCLOAK_GROUPS,
        ),
        get_group_by_path=Mock(
            side_effect=lambda path: {
                "id": "00000000-0000-0000-0000-000000000000",
                "name": path.strip("/"),
                "path": path,
                "attributes": {},
                "realmRoles": [],
                "clientRoles": TEST_KEYCLOAK_GROUP_CLIENT_ROLES[path],
                "subGroups": [],
            }
        ),
    ),
)
def test_cli_keycloak_listgroups_happy_path_from_env(_mock_keycloak_admin):
    result = run_cli_keycloak_listgroups(use_env=True)
    assert 0 == result.exit_code
    assert not result.exception

    parsed_result = parse_yaml(result.stdout, ["name", "roles"])

    parsed_result_dict = {item["name"]: item["roles"] for item in parsed_result}

    for group_path, expected_roles in TEST_KEYCLOAK_GROUP_CLIENT_ROLES.items():
        group_name = group_path.strip("/")
        assert group_name in parsed_result_dict
        assert (
            get_expected_roles(expected_roles, group_name)
            == parsed_result_dict[group_name]
        )


@patch(
    "keycloak.KeycloakAdmin",
    return_value=Mock(
        get_groups=Mock(
            side_effect=lambda: TEST_KEYCLOAK_GROUPS,
        ),
        get_group_by_path=Mock(
            side_effect=lambda path: {
                "id": "00000000-0000-0000-0000-000000000000",
                "name": path.strip("/"),
                "path": path,
                "attributes": {},
                "realmRoles": [],
                "clientRoles": TEST_KEYCLOAK_GROUP_CLIENT_ROLES[path],
                "subGroups": [],
            }
        ),
    ),
)
def test_cli_keycloak_listgroups_happy_path_from_config(_mock_keycloak_admin):
    result = run_cli_keycloak_listgroups(use_env=False)
    assert 0 == result.exit_code
    assert not result.exception

    parsed_result = parse_yaml(result.stdout, ["name", "roles"])

    parsed_result_dict = {item["name"]: item["roles"] for item in parsed_result}

    for group_path, expected_roles in TEST_KEYCLOAK_GROUP_CLIENT_ROLES.items():
        group_name = group_path.strip("/")
        assert group_name in parsed_result_dict
        assert (
            get_expected_roles(expected_roles, group_name)
            == parsed_result_dict[group_name]
        )


@patch(
    "keycloak.KeycloakAdmin.__init__",
    side_effect=keycloak.exceptions.KeycloakConnectionError(
        error_message="connection test"
    ),
)
def test_cli_keycloak_listusers_keycloak_connection_exception(_mock_keycloak_admin):
    result = run_cli_keycloak_listusers()

    assert 1 == result.exit_code
    assert result.exception
    assert "Failed to connect to Keycloak server: connection test" in str(
        result.exception
    )


@patch(
    "keycloak.KeycloakAdmin.__init__",
    side_effect=keycloak.exceptions.KeycloakAuthenticationError(
        error_message="auth test"
    ),
)
def test_cli_keycloak_listusers_keycloak_auth_exception(_mock_keycloak_admin):
    result = run_cli_keycloak_listusers()

    assert 1 == result.exit_code
    assert result.exception
    assert "Failed to connect to Keycloak server: auth test" in str(result.exception)


@patch(
    "keycloak.KeycloakAdmin",
    return_value=Mock(
        users_count=Mock(
            side_effect=keycloak.exceptions.KeycloakConnectionError(
                error_message="unhandled"
            )
        ),
    ),
)
def test_cli_keycloak_listusers_keycloak_unhandled_error(_mock_keycloak_admin):
    result = run_cli_keycloak_listusers()

    assert 1 == result.exit_code
    assert result.exception
    assert "unhandled" == str(result.exception)


def mock_api_post(admin_password: str, url: str, headers: Any, data: Any, verify: bool):
    response = Mock()
    if (
        url
        == f"{MOCK_KEYCLOAK_ENV['KEYCLOAK_SERVER_URL']}realms/master/protocol/openid-connect/token"
        and data["password"] == admin_password
    ):
        response.status_code = 200
        response.content = bytes(
            json.dumps({"access_token": TEST_ACCESS_TOKEN}), "UTF-8"
        )
    else:
        response.status_code = 403
    return response


def mock_api_request(
    access_token: str, method: str, url: str, headers: Any, verify: bool
):
    response = Mock()
    if (
        method == "GET"
        and url
        == f"{MOCK_KEYCLOAK_ENV['KEYCLOAK_SERVER_URL']}admin/realms/test-realm/users"
        and headers["Authorization"] == f"Bearer {access_token}"
    ):
        response.status_code = 200
        response.content = bytes(json.dumps(TEST_KEYCLOAK_USERS), "UTF-8")
    else:
        response.status_code = 403
    return response


@patch(
    "_nebari.keycloak.requests.post",
    side_effect=lambda url, headers, data, verify: mock_api_post(
        MOCK_KEYCLOAK_ENV["KEYCLOAK_ADMIN_PASSWORD"], url, headers, data, verify
    ),
)
@patch(
    "_nebari.keycloak.requests.request",
    side_effect=lambda method, url, headers, verify: mock_api_request(
        TEST_ACCESS_TOKEN, method, url, headers, verify
    ),
)
def test_cli_keycloak_exportusers_happy_path_from_env(
    _mock_requests_post, _mock_requests_request
):
    result = run_cli_keycloak_exportusers()

    assert 0 == result.exit_code
    assert not result.exception

    r = json.loads(result.stdout)
    assert "test-realm" == r["realm"]
    assert 3 == len(r["users"])
    assert "test-dev" == r["users"][0]["username"]


@patch(
    "_nebari.keycloak.requests.post",
    side_effect=lambda url, headers, data, verify: mock_api_post(
        MOCK_KEYCLOAK_ENV["KEYCLOAK_ADMIN_PASSWORD"], url, headers, data, verify
    ),
)
@patch(
    "_nebari.keycloak.requests.request",
    side_effect=lambda method, url, headers, verify: mock_api_request(
        TEST_ACCESS_TOKEN, method, url, headers, verify
    ),
)
def test_cli_keycloak_exportusers_happy_path_from_config(
    _mock_requests_post, _mock_requests_request
):
    result = run_cli_keycloak_exportusers(use_env=False)

    assert 0 == result.exit_code
    assert not result.exception

    r = json.loads(result.stdout)
    assert "test-realm" == r["realm"]
    assert 3 == len(r["users"])
    assert "test-dev" == r["users"][0]["username"]


@patch(
    "_nebari.keycloak.requests.post",
    side_effect=lambda url, headers, data, verify: mock_api_post(
        "invalid_admin_password", url, headers, data, verify
    ),
)
def test_cli_keycloak_exportusers_error_authentication(_mock_requests_post):
    result = run_cli_keycloak_exportusers()

    assert 1 == result.exit_code
    assert result.exception
    assert "Unable to retrieve Keycloak API token" in str(result.exception)
    assert "Status code: 403" in str(result.exception)


@patch(
    "_nebari.keycloak.requests.post",
    side_effect=lambda url, headers, data, verify: mock_api_post(
        MOCK_KEYCLOAK_ENV["KEYCLOAK_ADMIN_PASSWORD"], url, headers, data, verify
    ),
)
@patch(
    "_nebari.keycloak.requests.request",
    side_effect=lambda method, url, headers, verify: mock_api_request(
        "invalid_access_token", method, url, headers, verify
    ),
)
def test_cli_keycloak_exportusers_error_authorization(
    _mock_requests_post, _mock_requests_request
):
    result = run_cli_keycloak_exportusers()

    assert 1 == result.exit_code
    assert result.exception
    assert "Unable to communicate with Keycloak API" in str(result.exception)
    assert "Status code: 403" in str(result.exception)


@patch(
    "_nebari.keycloak.requests.post", side_effect=requests.exceptions.RequestException()
)
def test_cli_keycloak_exportusers_request_exception(_mock_requests_post):
    result = run_cli_keycloak_exportusers()

    assert 1 == result.exit_code
    assert result.exception


@patch("_nebari.keycloak.requests.post", side_effect=Exception())
def test_cli_keycloak_exportusers_unhandled_error(_mock_requests_post):
    result = run_cli_keycloak_exportusers()

    assert 1 == result.exit_code
    assert result.exception


def run_cli_keycloak(command: str, use_env: bool, extra_args: List[str] = []):
    with tempfile.TemporaryDirectory() as tmp:
        tmp_file = Path(tmp).resolve() / "nebari-config.yaml"
        assert tmp_file.exists() is False

        extra_config = (
            {
                "domain": TEST_DOMAIN,
                "security": {
                    "keycloak": {
                        "initial_root_password": MOCK_KEYCLOAK_ENV[
                            "KEYCLOAK_ADMIN_PASSWORD"
                        ]
                    }
                },
            }
            if not use_env
            else {}
        )
        config = {**{"project_name": "keycloak"}, **extra_config}
        with open(tmp_file.resolve(), "w") as f:
            yaml.dump(config, f)

        assert tmp_file.exists() is True

        app = create_cli()

        args = [
            "keycloak",
            command,
            "--config",
            tmp_file.resolve(),
        ] + extra_args

        env = MOCK_KEYCLOAK_ENV if use_env else {}
        result = runner.invoke(app, args=args, env=env)

        return result


def run_cli_keycloak_adduser(use_env: bool = True):
    username = TEST_KEYCLOAK_USERS[0]["username"]
    password = "test-password-123!"

    return run_cli_keycloak(
        "add-user",
        use_env=use_env,
        extra_args=[
            "--user",
            username,
            "--password",
            password,
        ],
    )


def run_cli_keycloak_listusers(use_env: bool = True):
    return run_cli_keycloak(
        "list-users",
        use_env=use_env,
    )


def run_cli_keycloak_listgroups(use_env: bool = True):
    return run_cli_keycloak(
        "list-groups",
        use_env=use_env,
    )


def run_cli_keycloak_exportusers(use_env: bool = True):
    return run_cli_keycloak(
        "export-users",
        use_env=use_env,
        extra_args=[
            "--realm",
            "test-realm",
        ],
    )



import inspect
import json

import kubernetes.client.models
from tornado import gen

kubernetes.client.models.V1EndpointPort = (
    kubernetes.client.models.CoreV1EndpointPort
)  # noqa: E402

from kubespawner import KubeSpawner  # noqa: E402

# conda-store default page size
DEFAULT_PAGE_SIZE_LIMIT = 100


@gen.coroutine
def get_username_hook(spawner):
    auth_state = yield spawner.user.get_auth_state()
    username = auth_state["oauth_user"]["preferred_username"]

    spawner.environment.update(
        {
            "PREFERRED_USERNAME": username,
        }
    )


def get_total_records(url: str, token: str) -> int:
    import urllib3

    http = urllib3.PoolManager()
    response = http.request("GET", url, headers={"Authorization": f"Bearer {token}"})
    decoded_response = json.loads(response.data.decode("UTF-8"))
    return decoded_response.get("count", 0)


def generate_paged_urls(base_url: str, total_records: int, page_size: int) -> list[str]:
    import math

    urls = []
    # pages starts at 1
    for page in range(1, math.ceil(total_records / page_size) + 1):
        urls.append(f"{base_url}?size={page_size}&page={page}")

    return urls


def get_scoped_token(
    conda_store_url: str,
    admin_token: str,
    name: str,
    groups: list[str] = None,
    admin: bool = False,
):
    """Generate a conda store auth token for a given set of groups. The
    token will only have `view` permissions for each group (regardless if
    the user has higher permissions associated with the group generally).

    If the user is an admin, then the token will have `view` permissions
    for all namespaces.

    By default, the user will have view permissions for the following
    groups:
    - `default/*`
    - `filesystem/*`
    - `nebari-git/*`
    - `global/*`

    Parameters
    ----------
    conda_store_url : str
        The URL of the Conda Store instance.
    admin_token : str
        The admin token used to authenticate with the Conda Store API.
    name : str
        The name of the user for whom the token will be generated (the token
        will have `view` permission to this namespace).
    groups : list[str], optional
        A list of group names that the user should have access to. If not provided,
        default roles will be assigned.
    admin : bool, optional
        Whether the user is an admin. If true, the token will have `view` permissions
        for all namespaces. Defaults to False.

    Returns
    -------
    str
        The generated token.
    """
    import urllib3

    token_endpoint = f"http://{conda_store_url}/conda-store/api/v1/token/"
    http = urllib3.PoolManager()

    # add default role bindings
    role_bindings = {
        "role_bindings": {
            f"{name}/*": ["viewer"],
            "default/*": ["viewer"],
            "filesystem/*": ["viewer"],
            "nebari-git/*": ["viewer"],
            "global/*": ["viewer"],
        }
    }

    # add role bindings for all the groups the user is part of
    if groups is not None:
        for group in groups:
            group = group.replace("/", "")
            role_bindings["role_bindings"][f"{group}/*"] = ["viewer"]

    # if the user is an admin, they can view all namespace + environments
    if admin:
        role_bindings["role_bindings"]["*/*"] = ["viewer"]

    encoded_body = json.dumps(role_bindings)

    # generate a token with with the generated role bindings
    token_response = http.request(
        "POST",
        str(token_endpoint),
        headers={
            "Authorization": f"Bearer {admin_token}",
            "Content-Type": "application/json",
        },
        body=encoded_body,
    )
    token_data = json.loads(token_response.data.decode("UTF-8"))
    return token_data.get("data", {}).get("token")


# TODO: this should get unit tests. Currently, since this is not a python module,
# adding tests in a traditional sense is not possible. See https://github.com/soapy1/nebari/tree/try-unit-test-spawner
# for a demo on one approach to adding test.
def get_conda_store_environments(user_info: dict):
    """Gets the conda-store environments for a given user using the v1 environment
    API.

    This scopes permissions for the given user using the `groups` field in the
    user_info dict. The user_info dict comes from Jupyter's user dict. The groups
    in this dict come from the keycloak groups, which include the list of conda-store
    namespaces the user has access to. For the purpose of this function, we can assume
    that if the user is part of the group, it at least has `view` permissions on the
    conda-store namespaces.

    Parameters
    ----------
    user_info : dict
        A dictionary containing user information originating from the JupyterHub
        user info. The scheme of the user info is:
        ```
        {
            "name": "string",
            "admin": true,
            "roles": [
                "string"
            ],
            "groups": [
                "string"
            ],
            "server": "string",
            "pending": "spawn",
            "last_activity": "2019-08-24T14:15:22Z",
            "servers": {
            },
            "auth_state": {}
        }
        ```

    Returns
    -------
    list[str]
        A list of all conda-store environments for the given user
    """
    import os

    import urllib3

    # Check for the environment variable `CONDA_STORE_API_PAGE_SIZE_LIMIT`. Fall
    # back to using the default page size limit if not set.
    page_size = os.environ.get(
        "CONDA_STORE_API_PAGE_SIZE_LIMIT", DEFAULT_PAGE_SIZE_LIMIT
    )

    external_url = z2jh.get_config("custom.conda-store-service-name")
    token = z2jh.get_config("custom.conda-store-jhub-apps-token")
    endpoint = "conda-store/api/v1/environment"

    base_url = f"http://{external_url}/{endpoint}/"

    groups = user_info["groups"]
    name = user_info["name"]
    admin = user_info["admin"]
    # get token with appropriate scope for the user making the request
    scoped_token = get_scoped_token(external_url, token, name, groups, admin)

    # get total number of records from the endpoint
    total_records = get_total_records(base_url, scoped_token)

    # will contain all the environment info returned from the api
    env_data = []

    # generate a list of urls to hit to build the response
    urls = generate_paged_urls(base_url, total_records, page_size)

    http = urllib3.PoolManager()

    # get content from urls
    for url in urls:
        response = http.request(
            "GET", url, headers={"Authorization": f"Bearer {scoped_token}"}
        )
        decoded_response = json.loads(response.data.decode("UTF-8"))
        env_data += decoded_response.get("data", [])

    # Filter and return conda environments for the user
    envs = [f"{env['namespace']['name']}-{env['name']}" for env in env_data]
    return envs


c.Spawner.pre_spawn_hook = get_username_hook

c.JupyterHub.allow_named_servers = False
c.JupyterHub.spawner_class = KubeSpawner

if z2jh.get_config("custom.jhub-apps-enabled"):
    from jhub_apps import theme_template_paths
    from jhub_apps.configuration import install_jhub_apps

    domain = z2jh.get_config("custom.external-url")
    hub_url = f"https://{domain}"
    c.JupyterHub.bind_url = hub_url
    c.JupyterHub.default_url = "/hub/home"
    c.Spawner.debug = True

    c.JAppsConfig.conda_envs = get_conda_store_environments
    c.JAppsConfig.jupyterhub_config_path = (
        "/usr/local/etc/jupyterhub/jupyterhub_config.py"
    )
    c.JAppsConfig.hub_host = "hub"
    c.JAppsConfig.service_workers = 4

    jhub_apps_overrides = json.loads(z2jh.get_config("custom.jhub-apps-overrides"))
    for config_key, config_value in jhub_apps_overrides.items():
        setattr(c.JAppsConfig, config_key, config_value)

    def service_for_jhub_apps(name, url):
        return {
            "name": name,
            "display": True,
            "info": {
                "name": name,
                "url": url,
                "external": True,
            },
        }

    c.JupyterHub.services.extend(
        [
            service_for_jhub_apps(name="Argo", url="/argo"),
            service_for_jhub_apps(name="Users", url="/auth/admin/nebari/console/"),
            service_for_jhub_apps(name="Environments", url="/conda-store"),
            service_for_jhub_apps(name="Monitoring", url="/monitoring"),
        ]
    )

    c.JupyterHub.template_paths = theme_template_paths

    kwargs = {}
    jhub_apps_signature = inspect.signature(install_jhub_apps)
    if "oauth_no_confirm" in jhub_apps_signature.parameters:
        kwargs["oauth_no_confirm"] = True

    c = install_jhub_apps(c, spawner_to_subclass=KubeSpawner, **kwargs)



import contextlib
import enum
import inspect
import os
import pathlib
import re
import sys
import tempfile
import warnings
from typing import Annotated, Any, Dict, List, Literal, Optional, Tuple, Type, Union

from pydantic import AfterValidator, ConfigDict, Field, field_validator, model_validator

from _nebari import constants
from _nebari.provider import opentofu
from _nebari.provider.cloud import amazon_web_services, azure_cloud, google_cloud
from _nebari.stages.base import NebariTerraformStage
from _nebari.stages.kubernetes_services import SharedFsEnum
from _nebari.stages.tf_objects import NebariTerraformState
from _nebari.utils import (
    AZURE_NODE_RESOURCE_GROUP_SUFFIX,
    construct_azure_resource_group_name,
    modified_environ,
)
from nebari import schema
from nebari.hookspecs import NebariStage, hookimpl


def get_kubeconfig_filename():
    return str(pathlib.Path(tempfile.gettempdir()) / "NEBARI_KUBECONFIG")


class LocalInputVars(schema.Base):
    kubeconfig_filename: str = get_kubeconfig_filename()
    kube_context: Optional[str] = None


class ExistingInputVars(schema.Base):
    kube_context: str


class NodeGroup(schema.Base):
    instance: str
    min_nodes: Annotated[int, Field(ge=0)] = 0
    max_nodes: Annotated[int, Field(ge=1)] = 1
    taints: Optional[List[schema.Taint]] = None

    @field_validator("taints", mode="before")
    def validate_taint_strings(cls, taints: list[Any]):
        if taints is None:
            return taints

        TAINT_STR_REGEX = re.compile(r"(\w+)=(\w+):(\w+)")
        return_value = []
        for taint in taints:
            if not isinstance(taint, str):
                return_value.append(taint)
            else:
                match = TAINT_STR_REGEX.match(taint)
                if not match:
                    raise ValueError(f"Invalid taint string: {taint}")
                key, taints, effect = match.groups()
                parsed_taint = schema.Taint(key=key, value=taints, effect=effect)
                return_value.append(parsed_taint)

        return return_value


DEFAULT_GENERAL_NODE_GROUP_TAINTS = []
DEFAULT_NODE_GROUP_TAINTS = [
    schema.Taint(key="dedicated", value="nebari", effect="NoSchedule")
]


def set_missing_taints_to_default_taints(node_groups: NodeGroup) -> NodeGroup:

    for node_group_name, node_group in node_groups.items():
        if node_group.taints is None:
            if node_group_name == "general":
                node_group.taints = DEFAULT_GENERAL_NODE_GROUP_TAINTS
            else:
                node_group.taints = DEFAULT_NODE_GROUP_TAINTS
    return node_groups


class GCPNodeGroupInputVars(schema.Base):
    name: str
    instance_type: str
    min_size: int
    max_size: int
    node_taints: List[dict]
    labels: Dict[str, str]
    preemptible: bool
    guest_accelerators: List["GCPGuestAccelerator"]

    @field_validator("node_taints", mode="before")
    def convert_taints(cls, value: Optional[List[schema.Taint]]):
        return [
            dict(
                key=taint.key,
                value=taint.value,
                effect={
                    schema.TaintEffectEnum.NoSchedule: "NO_SCHEDULE",
                    schema.TaintEffectEnum.PreferNoSchedule: "PREFER_NO_SCHEDULE",
                    schema.TaintEffectEnum.NoExecute: "NO_EXECUTE",
                }[taint.effect],
            )
            for taint in value
        ]


class GCPPrivateClusterConfig(schema.Base):
    enable_private_nodes: bool
    enable_private_endpoint: bool
    master_ipv4_cidr_block: str


@schema.yaml_object(schema.yaml)
class GCPNodeGroupImageTypeEnum(str, enum.Enum):
    UBUNTU_CONTAINERD = "UBUNTU_CONTAINERD"
    COS_CONTAINERD = "COS_CONTAINERD"

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_str(node.value)


class GCPInputVars(schema.Base):
    name: str
    environment: str
    region: str
    project_id: str
    availability_zones: List[str]
    node_groups: List[GCPNodeGroupInputVars]
    kubeconfig_filename: str = get_kubeconfig_filename()
    tags: List[str]
    kubernetes_version: str
    release_channel: str
    networking_mode: str
    network: str
    subnetwork: Optional[str] = None
    ip_allocation_policy: Optional[Dict[str, str]] = None
    master_authorized_networks_config: Optional[Dict[str, str]] = None
    private_cluster_config: Optional[GCPPrivateClusterConfig] = None
    node_group_image_type: GCPNodeGroupImageTypeEnum = None


class AzureNodeGroupInputVars(schema.Base):
    instance: str
    min_nodes: int
    max_nodes: int
    node_taints: list[str]

    @field_validator("node_taints", mode="before")
    def convert_taints(cls, value: Optional[List[schema.Taint]]):
        return [f"{taint.key}={taint.value}:{taint.effect.value}" for taint in value]


class AzureInputVars(schema.Base):
    name: str
    environment: str
    region: str
    authorized_ip_ranges: List[str] = ["0.0.0.0/0"]
    kubeconfig_filename: str = get_kubeconfig_filename()
    kubernetes_version: str
    node_groups: Dict[str, AzureNodeGroupInputVars]
    resource_group_name: str
    node_resource_group_name: str
    vnet_subnet_id: Optional[str] = None
    private_cluster_enabled: bool
    tags: Dict[str, str] = {}
    max_pods: Optional[int] = None
    network_profile: Optional[Dict[str, str]] = None
    azure_policy_enabled: Optional[bool] = None
    workload_identity_enabled: bool = False


class AWSAmiTypes(str, enum.Enum):
    AL2_x86_64 = "AL2_x86_64"
    AL2_x86_64_GPU = "AL2_x86_64_GPU"
    CUSTOM = "CUSTOM"


class AWSNodeLaunchTemplate(schema.Base):
    pre_bootstrap_command: Optional[str] = None
    ami_id: Optional[str] = None


class AWSNodeGroupInputVars(schema.Base):
    name: str
    instance_type: str
    gpu: bool = False
    min_size: int
    desired_size: int
    max_size: int
    single_subnet: bool
    permissions_boundary: Optional[str] = None
    ami_type: Optional[AWSAmiTypes] = None
    launch_template: Optional[AWSNodeLaunchTemplate] = None
    node_taints: list[dict]

    @field_validator("node_taints", mode="before")
    def convert_taints(cls, value: Optional[List[schema.Taint]]):
        return [
            dict(
                key=taint.key,
                value=taint.value,
                effect={
                    schema.TaintEffectEnum.NoSchedule: "NO_SCHEDULE",
                    schema.TaintEffectEnum.PreferNoSchedule: "PREFER_NO_SCHEDULE",
                    schema.TaintEffectEnum.NoExecute: "NO_EXECUTE",
                }[taint.effect],
            )
            for taint in value
        ]


def construct_aws_ami_type(
    gpu_enabled: bool, launch_template: AWSNodeLaunchTemplate
) -> str:
    """
    This function selects the Amazon Machine Image (AMI) type for AWS nodes by evaluating
    the provided parameters. The selection logic prioritizes the launch template over the
    GPU flag.

    Returns the AMI type (str) determined by the following rules:
        - Returns "CUSTOM" if a `launch_template` is provided and it includes a valid `ami_id`.
        - Returns "AL2_x86_64_GPU" if `gpu_enabled` is True and no valid
          `launch_template` is provided (None).
        - Returns "AL2_x86_64" as the default AMI type if `gpu_enabled` is False and no
          valid `launch_template` is provided (None).
    """

    if launch_template and getattr(launch_template, "ami_id", None):
        return "CUSTOM"

    if gpu_enabled:
        return "AL2_x86_64_GPU"

    return "AL2_x86_64"

    @field_validator("node_taints", mode="before")
    def convert_taints(cls, value: Optional[List[schema.Taint]]):
        return [
            dict(
                key=taint.key,
                value=taint.value,
                effect={
                    schema.TaintEffectEnum.NoSchedule: "NO_SCHEDULE",
                    schema.TaintEffectEnum.PreferNoSchedule: "PREFER_NO_SCHEDULE",
                    schema.TaintEffectEnum.NoExecute: "NO_EXECUTE",
                }[taint.effect],
            )
            for taint in value
        ]


class AWSInputVars(schema.Base):
    name: str
    environment: str
    existing_security_group_id: Optional[str] = None
    existing_subnet_ids: Optional[List[str]] = None
    region: str
    kubernetes_version: str
    eks_endpoint_access: Optional[
        Literal["private", "public", "public_and_private"]
    ] = "public"
    eks_kms_arn: Optional[str] = None
    eks_public_access_cidrs: Optional[List[str]] = ["0.0.0.0/0"]
    node_groups: List[AWSNodeGroupInputVars]
    availability_zones: List[str]
    vpc_cidr_block: str
    permissions_boundary: Optional[str] = None
    kubeconfig_filename: str = get_kubeconfig_filename()
    tags: Dict[str, str] = {}
    efs_enabled: bool


def _calculate_asg_node_group_map(config: schema.Main):
    if config.provider == schema.ProviderEnum.aws:
        return amazon_web_services.aws_get_asg_node_group_mapping(
            config.project_name,
            config.namespace,
            config.amazon_web_services.region,
        )
    else:
        return {}


def _calculate_node_groups(config: schema.Main):
    if config.provider == schema.ProviderEnum.aws:
        return {
            group: {"key": "eks.amazonaws.com/nodegroup", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config.provider == schema.ProviderEnum.gcp:
        return {
            group: {"key": "cloud.google.com/gke-nodepool", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config.provider == schema.ProviderEnum.azure:
        return {
            group: {"key": "azure-node-pool", "value": group}
            for group in ["general", "user", "worker"]
        }
    elif config.provider == schema.ProviderEnum.existing:
        return config.existing.model_dump()["node_selectors"]
    else:
        return config.local.model_dump()["node_selectors"]


def node_groups_to_dict(node_groups):
    return {ng_name: ng.model_dump() for ng_name, ng in node_groups.items()}


@contextlib.contextmanager
def kubernetes_provider_context(kubernetes_credentials: Dict[str, str]):
    credential_mapping = {
        "config_path": "KUBE_CONFIG_PATH",
        "config_context": "KUBE_CTX",
        "username": "KUBE_USER",
        "password": "KUBE_PASSWORD",
        "client_certificate": "KUBE_CLIENT_CERT_DATA",
        "client_key": "KUBE_CLIENT_KEY_DATA",
        "cluster_ca_certificate": "KUBE_CLUSTER_CA_CERT_DATA",
        "host": "KUBE_HOST",
        "token": "KUBE_TOKEN",
    }

    credentials = {
        credential_mapping[k]: v
        for k, v in kubernetes_credentials.items()
        if v is not None
    }
    with modified_environ(**credentials):
        yield


class KeyValueDict(schema.Base):
    key: str
    value: str


class GCPIPAllocationPolicy(schema.Base):
    cluster_secondary_range_name: str
    services_secondary_range_name: str
    cluster_ipv4_cidr_block: str
    services_ipv4_cidr_block: str


class GCPCIDRBlock(schema.Base):
    cidr_block: str
    display_name: str


class GCPMasterAuthorizedNetworksConfig(schema.Base):
    cidr_blocks: List[GCPCIDRBlock]


class GCPGuestAccelerator(schema.Base):
    """
    See general information regarding GPU support at:
    # TODO: replace with nebari.dev new URL
    https://docs.nebari.dev/en/stable/source/admin_guide/gpu.html?#add-gpu-node-group
    """

    name: str
    count: Annotated[int, Field(ge=1)] = 1


class GCPNodeGroup(NodeGroup):
    preemptible: bool = False
    labels: Dict[str, str] = {}
    guest_accelerators: List[GCPGuestAccelerator] = []


DEFAULT_GCP_NODE_GROUPS = {
    "general": GCPNodeGroup(
        instance="e2-standard-8",
        min_nodes=1,
        max_nodes=1,
    ),
    "user": GCPNodeGroup(
        instance="e2-standard-4",
        min_nodes=0,
        max_nodes=5,
    ),
    "worker": GCPNodeGroup(
        instance="e2-standard-4",
        min_nodes=0,
        max_nodes=5,
    ),
}


class GoogleCloudPlatformProvider(schema.Base):
    # If you pass a major and minor version without a patch version
    # yaml will pass it as a float, so we need to coerce it to a string
    model_config = ConfigDict(coerce_numbers_to_str=True)
    region: str
    project: str
    kubernetes_version: str
    availability_zones: Optional[List[str]] = []
    release_channel: str = constants.DEFAULT_GKE_RELEASE_CHANNEL
    node_groups: Annotated[
        Dict[str, GCPNodeGroup], AfterValidator(set_missing_taints_to_default_taints)
    ] = Field(DEFAULT_GCP_NODE_GROUPS, validate_default=True)
    tags: Optional[List[str]] = []
    networking_mode: str = "ROUTE"
    network: str = "default"
    subnetwork: Optional[Union[str, None]] = None
    ip_allocation_policy: Optional[Union[GCPIPAllocationPolicy, None]] = None
    master_authorized_networks_config: Optional[Union[GCPCIDRBlock, None]] = None
    private_cluster_config: Optional[Union[GCPPrivateClusterConfig, None]] = None

    @field_validator("kubernetes_version", mode="before")
    @classmethod
    def transform_version_to_str(cls, value) -> str:
        """Transforms the version to a string if it is not already."""
        return str(value)

    @model_validator(mode="before")
    @classmethod
    def _check_input(cls, data: Any) -> Any:
        available_regions = google_cloud.regions()
        if data["region"] not in available_regions:
            raise ValueError(
                f"Google Cloud region={data['region']} is not one of {available_regions}"
            )

        available_kubernetes_versions = google_cloud.kubernetes_versions(data["region"])
        if not any(
            v.startswith(str(data["kubernetes_version"]))
            for v in available_kubernetes_versions
        ):
            raise ValueError(
                f"\nInvalid `kubernetes-version` provided: {data['kubernetes_version']}.\nPlease select from one of the following supported Kubernetes versions: {available_kubernetes_versions} or omit flag to use latest Kubernetes version available."
            )  # noqa

        # check if instances are valid
        available_instances = google_cloud.instances(data["region"])
        if "node_groups" in data:
            for _, node_group in data["node_groups"].items():
                instance = (
                    node_group["instance"]
                    if hasattr(node_group, "__getitem__")
                    else node_group.instance
                )
                if instance not in available_instances:
                    raise ValueError(
                        f"Google Cloud Platform instance {instance} not one of available instance types={available_instances}"
                    )

        return data


class AzureNodeGroup(NodeGroup):
    pass


DEFAULT_AZURE_NODE_GROUPS = {
    "general": AzureNodeGroup(
        instance="Standard_D8_v3",
        min_nodes=1,
        max_nodes=1,
    ),
    "user": AzureNodeGroup(
        instance="Standard_D4_v3",
        min_nodes=0,
        max_nodes=5,
    ),
    "worker": AzureNodeGroup(
        instance="Standard_D4_v3",
        min_nodes=0,
        max_nodes=5,
    ),
}


class AzureProvider(schema.Base):
    region: str
    kubernetes_version: Optional[str] = None
    storage_account_postfix: str
    authorized_ip_ranges: Optional[List[str]] = ["0.0.0.0/0"]
    resource_group_name: Optional[str] = None
    node_groups: Annotated[
        Dict[str, AzureNodeGroup], AfterValidator(set_missing_taints_to_default_taints)
    ] = Field(DEFAULT_AZURE_NODE_GROUPS, validate_default=True)
    storage_account_postfix: str
    vnet_subnet_id: Optional[str] = None
    private_cluster_enabled: bool = False
    resource_group_name: Optional[str] = None
    tags: Optional[Dict[str, str]] = {}
    network_profile: Optional[Dict[str, str]] = None
    max_pods: Optional[int] = None
    workload_identity_enabled: bool = False
    azure_policy_enabled: Optional[bool] = None

    @model_validator(mode="before")
    @classmethod
    def _check_credentials(cls, data: Any) -> Any:
        azure_cloud.check_credentials()
        return data

    @field_validator("kubernetes_version")
    @classmethod
    def _validate_kubernetes_version(cls, value: Optional[str]) -> str:
        available_kubernetes_versions = azure_cloud.kubernetes_versions()
        if value is None:
            value = available_kubernetes_versions[-1]
        elif value not in available_kubernetes_versions:
            raise ValueError(
                f"\nInvalid `kubernetes-version` provided: {value}.\nPlease select from one of the following supported Kubernetes versions: {available_kubernetes_versions} or omit flag to use latest Kubernetes version available."
            )
        return value

    @field_validator("resource_group_name")
    @classmethod
    def _validate_resource_group_name(cls, value):
        if value is None:
            return value
        length = len(value) + len(AZURE_NODE_RESOURCE_GROUP_SUFFIX)
        if length < 1 or length > 90:
            raise ValueError(
                f"Azure Resource Group name must be between 1 and 90 characters long, when combined with the suffix `{AZURE_NODE_RESOURCE_GROUP_SUFFIX}`."
            )
        if not re.match(r"^[\w\-\.\(\)]+$", value):
            raise ValueError(
                "Azure Resource Group name can only contain alphanumerics, underscores, parentheses, hyphens, and periods."
            )
        if value[-1] == ".":
            raise ValueError("Azure Resource Group name can't end with a period.")

        return value

    @field_validator("tags")
    @classmethod
    def _validate_tags(cls, value: Optional[Dict[str, str]]) -> Dict[str, str]:
        return value if value is None else azure_cloud.validate_tags(value)


class AWSNodeGroup(NodeGroup):
    gpu: bool = False
    single_subnet: bool = False
    permissions_boundary: Optional[str] = None
    # Disabled as part of 2024.11.1 until #2832 is resolved
    # launch_template: Optional[AWSNodeLaunchTemplate] = None

    @model_validator(mode="before")
    def check_launch_template(cls, values):
        if "launch_template" in values:
            raise ValueError(
                "The 'launch_template' field is currently unavailable and has been removed from the configuration schema.\nPlease omit this field until it is reintroduced in a future update.",
            )
        return values


DEFAULT_AWS_NODE_GROUPS = {
    "general": AWSNodeGroup(
        instance="m5.2xlarge",
        min_nodes=1,
        max_nodes=1,
    ),
    "user": AWSNodeGroup(
        instance="m5.xlarge",
        min_nodes=0,
        max_nodes=5,
        single_subnet=False,
    ),
    "worker": AWSNodeGroup(
        instance="m5.xlarge",
        min_nodes=0,
        max_nodes=5,
        single_subnet=False,
    ),
}


class AmazonWebServicesProvider(schema.Base):
    region: str
    kubernetes_version: str
    availability_zones: Optional[List[str]]
    node_groups: Annotated[
        Dict[str, AWSNodeGroup], AfterValidator(set_missing_taints_to_default_taints)
    ] = Field(DEFAULT_AWS_NODE_GROUPS, validate_default=True)
    eks_endpoint_access: Optional[
        Literal["private", "public", "public_and_private"]
    ] = "public"
    eks_public_access_cidrs: Optional[List[str]] = ["0.0.0.0/0"]
    eks_kms_arn: Optional[str] = None
    existing_subnet_ids: Optional[List[str]] = None
    existing_security_group_id: Optional[str] = None
    vpc_cidr_block: str = "10.10.0.0/16"
    permissions_boundary: Optional[str] = None
    tags: Optional[Dict[str, str]] = {}

    @model_validator(mode="before")
    @classmethod
    def _check_input(cls, data: Any) -> Any:
        amazon_web_services.check_credentials()

        # check if region is valid
        available_regions = amazon_web_services.regions(data["region"])
        if data["region"] not in available_regions:
            raise ValueError(
                f"Amazon Web Services region={data['region']} is not one of {available_regions}"
            )

        # check if kubernetes version is valid
        available_kubernetes_versions = amazon_web_services.kubernetes_versions(
            data["region"]
        )
        if len(available_kubernetes_versions) == 0:
            raise ValueError("Request to AWS for available Kubernetes versions failed.")
        if data["kubernetes_version"] is None:
            data["kubernetes_version"] = available_kubernetes_versions[-1]
        elif data["kubernetes_version"] not in available_kubernetes_versions:
            raise ValueError(
                f"\nInvalid `kubernetes-version` provided: {data['kubernetes_version']}.\nPlease select from one of the following supported Kubernetes versions: {available_kubernetes_versions} or omit flag to use latest Kubernetes version available."
            )

        # check if availability zones are valid
        available_zones = amazon_web_services.zones(data["region"])
        if "availability_zones" not in data:
            data["availability_zones"] = list(sorted(available_zones))[:2]
        else:
            for zone in data["availability_zones"]:
                if zone not in available_zones:
                    raise ValueError(
                        f"Amazon Web Services availability zone={zone} is not one of {available_zones}"
                    )

        # check if instances are valid
        available_instances = amazon_web_services.instances(data["region"])
        if "node_groups" in data:
            for _, node_group in data["node_groups"].items():
                instance = (
                    node_group["instance"]
                    if hasattr(node_group, "__getitem__")
                    else node_group.instance
                )
                if instance not in available_instances:
                    raise ValueError(
                        f"Amazon Web Services instance {node_group.instance} not one of available instance types={available_instances}"
                    )

        # check if kms key is valid
        available_kms_keys = amazon_web_services.kms_key_arns(data["region"])
        if "eks_kms_arn" in data and data["eks_kms_arn"] is not None:
            key_id = [
                id for id in available_kms_keys.keys() if id in data["eks_kms_arn"]
            ]
            # Raise error if key_id is not found in available_kms_keys
            if (
                len(key_id) != 1
                or available_kms_keys[key_id[0]].Arn != data["eks_kms_arn"]
            ):
                raise ValueError(
                    f"Amazon Web Services KMS Key with ARN {data['eks_kms_arn']} not one of available/enabled keys={ [v.Arn for v in available_kms_keys.values() if v.KeyManager == 'CUSTOMER' and v.KeySpec == 'SYMMETRIC_DEFAULT']}"
                )
            key_id = key_id[0]
            # Raise error if key is not a customer managed key
            if available_kms_keys[key_id].KeyManager != "CUSTOMER":
                raise ValueError(
                    f"Amazon Web Services KMS Key with ID {key_id} is not a customer managed key"
                )
            # Symmetric KMS keys with Encrypt and decrypt key-usage have the SYMMETRIC_DEFAULT key-spec
            # EKS cluster encryption requires a Symmetric key that is set to encrypt and decrypt data
            if available_kms_keys[key_id].KeySpec != "SYMMETRIC_DEFAULT":
                if available_kms_keys[key_id].KeyUsage == "GENERATE_VERIFY_MAC":
                    raise ValueError(
                        f"Amazon Web Services KMS Key with ID {key_id} does not have KeyUsage set to 'Encrypt and decrypt' data"
                    )
                elif available_kms_keys[key_id].KeyUsage != "ENCRYPT_DECRYPT":
                    raise ValueError(
                        f"Amazon Web Services KMS Key with ID {key_id} is not of type Symmetric, and KeyUsage not set to 'Encrypt and decrypt' data"
                    )
                else:
                    raise ValueError(
                        f"Amazon Web Services KMS Key with ID {key_id} is not of type Symmetric"
                    )

        return data


class LocalProvider(schema.Base):
    kube_context: Optional[str] = None
    node_selectors: Dict[str, KeyValueDict] = {
        "general": KeyValueDict(key="kubernetes.io/os", value="linux"),
        "user": KeyValueDict(key="kubernetes.io/os", value="linux"),
        "worker": KeyValueDict(key="kubernetes.io/os", value="linux"),
    }


class ExistingProvider(schema.Base):
    kube_context: Optional[str] = None
    node_selectors: Dict[str, KeyValueDict] = {
        "general": KeyValueDict(key="kubernetes.io/os", value="linux"),
        "user": KeyValueDict(key="kubernetes.io/os", value="linux"),
        "worker": KeyValueDict(key="kubernetes.io/os", value="linux"),
    }


provider_enum_model_map = {
    schema.ProviderEnum.local: LocalProvider,
    schema.ProviderEnum.existing: ExistingProvider,
    schema.ProviderEnum.gcp: GoogleCloudPlatformProvider,
    schema.ProviderEnum.aws: AmazonWebServicesProvider,
    schema.ProviderEnum.azure: AzureProvider,
}

provider_name_abbreviation_map: Dict[str, str] = {
    value: key.value for key, value in schema.provider_enum_name_map.items()
}

provider_enum_default_node_groups_map: Dict[schema.ProviderEnum, Any] = {
    schema.ProviderEnum.gcp: node_groups_to_dict(DEFAULT_GCP_NODE_GROUPS),
    schema.ProviderEnum.aws: node_groups_to_dict(DEFAULT_AWS_NODE_GROUPS),
    schema.ProviderEnum.azure: node_groups_to_dict(DEFAULT_AZURE_NODE_GROUPS),
}


class InputSchema(schema.Base):
    local: Optional[LocalProvider] = None
    existing: Optional[ExistingProvider] = None
    google_cloud_platform: Optional[GoogleCloudPlatformProvider] = None
    amazon_web_services: Optional[AmazonWebServicesProvider] = None
    azure: Optional[AzureProvider] = None

    @model_validator(mode="before")
    @classmethod
    def check_provider(cls, data: Any) -> Any:
        if "provider" in data:
            provider: str = data["provider"]
            if hasattr(schema.ProviderEnum, provider):
                # TODO: all cloud providers has required fields, but local and existing don't.
                #  And there is no way to initialize a model without user input here.
                #  We preserve the original behavior here, but we should find a better way to do this.
                if provider in ["local", "existing"] and provider not in data:
                    data[provider] = provider_enum_model_map[provider]()
            else:
                # if the provider field is invalid, it won't be set when this validator is called
                # so we need to check for it explicitly here, and set mode to "before"
                # TODO: this is a workaround, check if there is a better way to do this in Pydantic v2
                raise ValueError(
                    f"'{provider}' is not a valid enumeration member; permitted: local, existing, aws, gcp, azure"
                )
            set_providers = {
                provider
                for provider in provider_name_abbreviation_map.keys()
                if provider in data and data[provider]
            }
            expected_provider_config = schema.provider_enum_name_map[provider]
            extra_provider_config = set_providers - {expected_provider_config}
            if extra_provider_config:
                warnings.warn(
                    f"Provider is set to {getattr(provider, 'value', provider)},  but configuration defined for other providers: {extra_provider_config}"
                )

        else:
            set_providers = [
                provider
                for provider in provider_name_abbreviation_map.keys()
                if provider in data
            ]
            num_providers = len(set_providers)
            if num_providers > 1:
                raise ValueError(f"Multiple providers set: {set_providers}")
            elif num_providers == 1:
                data["provider"] = provider_name_abbreviation_map[set_providers[0]]
            elif num_providers == 0:
                data["provider"] = schema.ProviderEnum.local.value

        return data


class NodeSelectorKeyValue(schema.Base):
    key: str
    value: str


class KubernetesCredentials(schema.Base):
    host: str
    cluster_ca_certificate: str
    token: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    client_certificate: Optional[str] = None
    client_key: Optional[str] = None
    config_path: Optional[str] = None
    config_context: Optional[str] = None


class OutputSchema(schema.Base):
    node_selectors: Dict[str, NodeSelectorKeyValue]
    kubernetes_credentials: KubernetesCredentials
    kubeconfig_filename: str
    nfs_endpoint: Optional[str] = None


class KubernetesInfrastructureStage(NebariTerraformStage):
    """Generalized method to provision infrastructure.

    After successful deployment the following properties are set on
    `stage_outputs[directory]`.
      - `kubernetes_credentials` which are sufficient credentials to
        connect with the kubernetes provider
      - `kubeconfig_filename` which is a path to a kubeconfig that can
        be used to connect to a kubernetes cluster
      - at least one node running such that resources in the
        node_group.general can be scheduled

    At a high level this stage is expected to provision a kubernetes
    cluster on a given provider.
    """

    name = "02-infrastructure"
    priority = 20

    input_schema = InputSchema
    output_schema = OutputSchema

    @property
    def template_directory(self):
        return (
            pathlib.Path(inspect.getfile(self.__class__)).parent
            / "template"
            / self.config.provider.value
        )

    @property
    def stage_prefix(self):
        return pathlib.Path("stages") / self.name / self.config.provider.value

    def state_imports(self) -> List[Tuple[str, str]]:
        if self.config.provider == schema.ProviderEnum.azure:
            if self.config.azure.resource_group_name is None:
                return []

            subscription_id = os.environ["ARM_SUBSCRIPTION_ID"]
            resource_group_name = construct_azure_resource_group_name(
                project_name=self.config.project_name,
                namespace=self.config.namespace,
                base_resource_group_name=self.config.azure.resource_group_name,
            )
            resource_url = (
                f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}"
            )
            return [
                (
                    "azurerm_resource_group.resource_group",
                    resource_url,
                )
            ]

    def tf_objects(self) -> List[Dict]:
        if self.config.provider == schema.ProviderEnum.gcp:
            return [
                opentofu.Provider(
                    "google",
                    project=self.config.google_cloud_platform.project,
                    region=self.config.google_cloud_platform.region,
                ),
                NebariTerraformState(self.name, self.config),
            ]
        elif self.config.provider == schema.ProviderEnum.azure:
            return [
                NebariTerraformState(self.name, self.config),
            ]
        elif self.config.provider == schema.ProviderEnum.aws:
            return [
                opentofu.Provider("aws", region=self.config.amazon_web_services.region),
                NebariTerraformState(self.name, self.config),
            ]
        else:
            return []

    def input_vars(self, stage_outputs: Dict[str, Dict[str, Any]]):
        if self.config.provider == schema.ProviderEnum.local:
            return LocalInputVars(
                kube_context=self.config.local.kube_context
            ).model_dump()
        elif self.config.provider == schema.ProviderEnum.existing:
            return ExistingInputVars(
                kube_context=self.config.existing.kube_context
            ).model_dump()
        elif self.config.provider == schema.ProviderEnum.gcp:
            return GCPInputVars(
                name=self.config.escaped_project_name,
                environment=self.config.namespace,
                region=self.config.google_cloud_platform.region,
                project_id=self.config.google_cloud_platform.project,
                availability_zones=self.config.google_cloud_platform.availability_zones,
                node_groups=[
                    GCPNodeGroupInputVars(
                        name=name,
                        labels=node_group.labels,
                        instance_type=node_group.instance,
                        min_size=node_group.min_nodes,
                        max_size=node_group.max_nodes,
                        node_taints=node_group.taints,
                        preemptible=node_group.preemptible,
                        guest_accelerators=node_group.guest_accelerators,
                    )
                    for name, node_group in self.config.google_cloud_platform.node_groups.items()
                ],
                tags=self.config.google_cloud_platform.tags,
                kubernetes_version=self.config.google_cloud_platform.kubernetes_version,
                release_channel=self.config.google_cloud_platform.release_channel,
                networking_mode=self.config.google_cloud_platform.networking_mode,
                network=self.config.google_cloud_platform.network,
                subnetwork=self.config.google_cloud_platform.subnetwork,
                ip_allocation_policy=self.config.google_cloud_platform.ip_allocation_policy,
                master_authorized_networks_config=self.config.google_cloud_platform.master_authorized_networks_config,
                private_cluster_config=self.config.google_cloud_platform.private_cluster_config,
                node_group_image_type=(
                    GCPNodeGroupImageTypeEnum.UBUNTU_CONTAINERD
                    if self.config.storage.type == SharedFsEnum.cephfs
                    else GCPNodeGroupImageTypeEnum.COS_CONTAINERD
                ),
            ).model_dump()
        elif self.config.provider == schema.ProviderEnum.azure:
            return AzureInputVars(
                name=self.config.escaped_project_name,
                environment=self.config.namespace,
                region=self.config.azure.region,
                kubernetes_version=self.config.azure.kubernetes_version,
                authorized_ip_ranges=self.config.azure.authorized_ip_ranges,
                node_groups={
                    name: AzureNodeGroupInputVars(
                        instance=node_group.instance,
                        min_nodes=node_group.min_nodes,
                        max_nodes=node_group.max_nodes,
                        node_taints=node_group.taints,
                    )
                    for name, node_group in self.config.azure.node_groups.items()
                },
                resource_group_name=construct_azure_resource_group_name(
                    project_name=self.config.project_name,
                    namespace=self.config.namespace,
                    base_resource_group_name=self.config.azure.resource_group_name,
                ),
                node_resource_group_name=construct_azure_resource_group_name(
                    project_name=self.config.project_name,
                    namespace=self.config.namespace,
                    base_resource_group_name=self.config.azure.resource_group_name,
                    suffix=AZURE_NODE_RESOURCE_GROUP_SUFFIX,
                ),
                vnet_subnet_id=self.config.azure.vnet_subnet_id,
                private_cluster_enabled=self.config.azure.private_cluster_enabled,
                tags=self.config.azure.tags,
                network_profile=self.config.azure.network_profile,
                max_pods=self.config.azure.max_pods,
                workload_identity_enabled=self.config.azure.workload_identity_enabled,
                azure_policy_enabled=self.config.azure.azure_policy_enabled,
            ).model_dump()
        elif self.config.provider == schema.ProviderEnum.aws:
            return AWSInputVars(
                name=self.config.escaped_project_name,
                environment=self.config.namespace,
                eks_endpoint_access=self.config.amazon_web_services.eks_endpoint_access,
                eks_public_access_cidrs=self.config.amazon_web_services.eks_public_access_cidrs,
                eks_kms_arn=self.config.amazon_web_services.eks_kms_arn,
                existing_subnet_ids=self.config.amazon_web_services.existing_subnet_ids,
                existing_security_group_id=self.config.amazon_web_services.existing_security_group_id,
                region=self.config.amazon_web_services.region,
                kubernetes_version=self.config.amazon_web_services.kubernetes_version,
                node_groups=[
                    AWSNodeGroupInputVars(
                        name=name,
                        instance_type=node_group.instance,
                        gpu=node_group.gpu,
                        min_size=node_group.min_nodes,
                        desired_size=node_group.min_nodes,
                        max_size=node_group.max_nodes,
                        single_subnet=node_group.single_subnet,
                        permissions_boundary=node_group.permissions_boundary,
                        launch_template=None,
                        node_taints=node_group.taints,
                        ami_type=construct_aws_ami_type(
                            gpu_enabled=node_group.gpu,
                            launch_template=None,
                        ),
                    )
                    for name, node_group in self.config.amazon_web_services.node_groups.items()
                ],
                availability_zones=self.config.amazon_web_services.availability_zones,
                vpc_cidr_block=self.config.amazon_web_services.vpc_cidr_block,
                permissions_boundary=self.config.amazon_web_services.permissions_boundary,
                tags=self.config.amazon_web_services.tags,
                efs_enabled=self.config.storage.type == SharedFsEnum.efs,
            ).model_dump()
        else:
            raise ValueError(f"Unknown provider: {self.config.provider}")

    def check(
        self, stage_outputs: Dict[str, Dict[str, Any]], disable_prompt: bool = False
    ):
        from kubernetes import client, config
        from kubernetes.client.rest import ApiException

        config.load_kube_config(
            config_file=stage_outputs["stages/02-infrastructure"][
                "kubeconfig_filename"
            ]["value"]
        )

        try:
            api_instance = client.CoreV1Api()
            result = api_instance.list_namespace()
        except ApiException:
            print(
                f"ERROR: After stage={self.name} unable to connect to kubernetes cluster"
            )
            sys.exit(1)

        if len(result.items) < 1:
            print(
                f"ERROR: After stage={self.name} no nodes provisioned within kubernetes cluster"
            )
            sys.exit(1)

        print(f"After stage={self.name} kubernetes cluster successfully provisioned")

    def set_outputs(
        self, stage_outputs: Dict[str, Dict[str, Any]], outputs: Dict[str, Any]
    ):
        outputs["node_selectors"] = _calculate_node_groups(self.config)
        super().set_outputs(stage_outputs, outputs)

    @contextlib.contextmanager
    def post_deploy(
        self, stage_outputs: Dict[str, Dict[str, Any]], disable_prompt: bool = False
    ):
        asg_node_group_map = _calculate_asg_node_group_map(self.config)
        if asg_node_group_map:
            amazon_web_services.set_asg_tags(
                asg_node_group_map, self.config.amazon_web_services.region
            )

    @contextlib.contextmanager
    def deploy(
        self, stage_outputs: Dict[str, Dict[str, Any]], disable_prompt: bool = False
    ):
        with super().deploy(stage_outputs, disable_prompt):
            with kubernetes_provider_context(
                stage_outputs["stages/" + self.name]["kubernetes_credentials"]["value"]
            ):
                yield

    @contextlib.contextmanager
    def destroy(
        self, stage_outputs: Dict[str, Dict[str, Any]], status: Dict[str, bool]
    ):
        with super().destroy(stage_outputs, status):
            with kubernetes_provider_context(
                stage_outputs["stages/" + self.name]["kubernetes_credentials"]["value"]
            ):
                yield


@hookimpl
def nebari_stage() -> List[Type[NebariStage]]:
    return [KubernetesInfrastructureStage]



import json
import logging
import os
from functools import wraps
from typing import Any, Callable, Dict, List
from urllib.parse import urljoin

import keycloak
import requests
import rich
import urllib3

from _nebari.stages.kubernetes_ingress import CertificateEnum
from nebari import schema

logger = logging.getLogger(__name__)


def do_keycloak() -> Callable:
    """
    A simple decorator *factory* to wrap functions that interact with Keycloak. Injects
    the config parameter depending on the function signature and expects the presence of
    the keycloak_admin parameter in the wrapped function.

    Usage:
        @do_keycloak()
        def keycloak_list_users(keycloak_admin):
            ...
        user = keycloak_list_users(config)
    """

    def _decorator(func: Callable) -> Callable:
        @wraps(func)
        def _wrapper(config, *args, **kwargs):
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

            keycloak_admin = get_keycloak_admin_from_config(config=config)

            return func(keycloak_admin, *args, **kwargs)

        return _wrapper

    return _decorator


@do_keycloak()
def create_user(
    keycloak_admin: keycloak.KeycloakAdmin,
    username: str,
    password: str = None,
    groups: list = None,
    email=None,
    domain=None,
    enabled=True,
):
    payload = {
        "username": username,
        "groups": groups or ["/developer"],
        "email": email or f"{username}@{domain or 'example.com'}",
        "enabled": enabled,
    }
    if password:
        payload["credentials"] = [
            {"type": "password", "value": password, "temporary": False}
        ]
    if groups:
        _available_group_names = [_["name"] for _ in keycloak_admin.get_groups()]
        if not all([_ in _available_group_names for _ in groups]):
            raise ValueError(f"Keycloak group(s) not found: {groups}")
    else:
        rich.print(
            f"Creating user=[green]{username}[/green] without password (none supplied)"
        )
    user = keycloak_admin.create_user(payload)
    rich.print(f"Created user=[green]{username}[/green]")
    return user


@do_keycloak()
def list_users(keycloak_admin: keycloak.KeycloakAdmin) -> Dict[str, Any]:
    """
    Return a list of user info from Keycloak, including username, email, and groups.
    """
    return {
        "count": keycloak_admin.users_count(),
        "users": [
            {
                "username": user["username"],
                "email": user["email"],
                "groups": [
                    group["name"]
                    for group in keycloak_admin.get_user_groups(user["id"])
                ],
            }
            for user in keycloak_admin.get_users()
        ],
    }


def get_expected_roles(expected_roles, group_name):
    return [
        role
        for client, roles in expected_roles.items()
        for role in roles
        if not (client == "realm-management" and group_name != "superadmin")
    ]


@do_keycloak()
def list_groups(keycloak_admin: keycloak.KeycloakAdmin) -> List[Dict[str, Any]]:
    """Return a list of group info from Keycloak, excluding 'realm-management' roles for all groups except 'superadmin'."""
    groups_data = []
    for group in keycloak_admin.get_groups():
        attrs = keycloak_admin.get_group_by_path(group["path"])
        client_roles = attrs.get("clientRoles", {})

        # Collect roles except 'realm-management' (unless group is 'superadmin')
        roles = get_expected_roles(client_roles, group_name=group["name"])

        groups_data.append(
            {
                "name": group["name"],
                "roles": roles,
            }
        )

    return groups_data


def get_keycloak_admin(server_url, username, password, verify=False):
    try:
        keycloak_admin = keycloak.KeycloakAdmin(
            server_url=server_url,
            username=username,
            password=password,
            realm_name=os.environ.get("KEYCLOAK_REALM", "nebari"),
            user_realm_name="master",
            auto_refresh_token=("get", "put", "post", "delete"),
            verify=verify,
        )
    except (
        keycloak.exceptions.KeycloakConnectionError,
        keycloak.exceptions.KeycloakAuthenticationError,
    ) as e:
        raise ValueError(f"Failed to connect to Keycloak server: {e}")

    return keycloak_admin


def get_keycloak_admin_from_config(config: schema.Main):
    keycloak_server_url = os.environ.get(
        "KEYCLOAK_SERVER_URL", f"https://{config.domain}/auth/"
    )

    keycloak_username = os.environ.get("KEYCLOAK_ADMIN_USERNAME", "root")
    keycloak_password = os.environ.get(
        "KEYCLOAK_ADMIN_PASSWORD", config.security.keycloak.initial_root_password
    )

    should_verify_tls = config.certificate.type != CertificateEnum.selfsigned

    return get_keycloak_admin(
        server_url=keycloak_server_url,
        username=keycloak_username,
        password=keycloak_password,
        verify=should_verify_tls,
    )


def keycloak_rest_api_call(config: schema.Main = None, request: str = None):
    """Communicate directly with the Keycloak REST API by passing it a request"""
    keycloak_server_url = os.environ.get(
        "KEYCLOAK_SERVER_URL", f"https://{config.domain}/auth/"
    )

    keycloak_admin_username = os.environ.get("KEYCLOAK_ADMIN_USERNAME", "root")
    keycloak_admin_password = os.environ.get(
        "KEYCLOAK_ADMIN_PASSWORD",
        config.security.keycloak.initial_root_password,
    )

    try:
        # Get `token` to interact with Keycloak Admin
        url = urljoin(
            keycloak_server_url, "realms/master/protocol/openid-connect/token"
        )
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {
            "username": keycloak_admin_username,
            "password": keycloak_admin_password,
            "grant_type": "password",
            "client_id": "admin-cli",
        }

        response = requests.post(
            url=url,
            headers=headers,
            data=data,
            verify=False,
        )

        if response.status_code == 200:
            token = json.loads(response.content.decode())["access_token"]
        else:
            raise ValueError(
                f"Unable to retrieve Keycloak API token. Status code: {response.status_code}"
            )

        # Send request to Keycloak REST API
        method, endpoint = request.split()
        url = urljoin(
            urljoin(keycloak_server_url, "admin/realms/"), endpoint.lstrip("/")
        )
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {token}",
        }

        response = requests.request(
            method=method, url=url, headers=headers, verify=False
        )

        if response.status_code == 200:
            content = json.loads(response.content.decode())
            return content
        else:
            raise ValueError(
                f"Unable to communicate with Keycloak API. Status code: {response.status_code}"
            )

    except requests.exceptions.RequestException as e:
        raise e


def export_keycloak_users(config: schema.Main, realm: str):
    request = f"GET /{realm}/users"

    users = keycloak_rest_api_call(config, request=request)

    return {
        "realm": realm,
        "users": users,
    }
